class ModelPerformance():
    score = None
    model_name = None
    RMSE = None
    total_columns = None
    scaler_type = None
    selected_features = None
    num_of_clusters_if_used = None
    accuracy = None
    f1_score = None

    def __init__(self,score,model_name,RMSE = None,total_columns = 0,scaler_type = None,selected_features = None,num_of_clusters = None, accuracy = None, f1_score = None) -> None:
        self.model_name = model_name
        self.score = score
        self.RMSE = RMSE
        self.total_columns = total_columns
        self.scaler_type = scaler_type
        self.selected_features = selected_features
        self.num_of_clusters_if_used = num_of_clusters
        self.accuracy = accuracy
        self.f1_score = f1_score

    def to_dict(self):
        return vars(self)
 


class NeuralNetwork_BayesianOptimization_Params():
    neurons_min_max = None
    activation_min_max = None
    optimizer_min_max = None
    learning_rate_min_max = None
    batch_size_min_max = None
    epochs_min_max = None
    normalization_min_max = None
    dropout_rate_min_max = None
    dropout_min_max = None
    hidden_layers_min_max = None

    def __init__(self
                 ,neurons_min_max = (32, 128)
                 ,learning_rate_min_max = (0.001, .01)
                 ,batch_size_min_max = (32, 64)
                 ,epochs_min_max = (50, 100)
                 ,normalization_min_max = (0,1)
                 ,dropout_rate_min_max = (0.2,0.6)
                 ,hidden_layers_min_max = (1,2)
                 ,dropout_min_max = (0,1)
                 ,activation_min_max = (0, 9) ):
        self.neurons_min_max = neurons_min_max
        self.activation_min_max = activation_min_max
        self.optimizer_min_max = (0,6)
        self.learning_rate_min_max = learning_rate_min_max
        self.batch_size_min_max = batch_size_min_max
        self.epochs_min_max = epochs_min_max
        self.normalization_min_max = normalization_min_max
        self.dropout_min_max = dropout_min_max
        self.dropout_rate_min_max = dropout_rate_min_max
        self.hidden_layers_min_max = hidden_layers_min_max
  


from auto_ml_kinder.classes.Functions import Functions
import pandas as pd
from auto_ml_kinder.classes.ModelMeta import ModelMeta
from auto_ml_kinder.classes.NeuralNetwork_BayesianOptimization import NeuralNetwork_BayesianOptimization
from auto_ml_kinder.classes.ModelAndParamRegression import ModelAndParamRegression
from auto_ml_kinder.classes.ModelPower import ModelPower
from auto_ml_kinder.classes.ModelAndParamClassifiction import ModelAndParamClassifiction
from auto_ml_kinder.classes.ModelPerformance import ModelPerformance
from auto_ml_kinder.classes.NeuralNetwork_BayesianOptimization_Params import NeuralNetwork_BayesianOptimization_Params
from auto_ml_kinder.classes.NeuralNetwork_Regression import NeuralNetwork_Regression
from auto_ml_kinder.classes.NeuralNetwork_Classification import NeuralNetwork_Classification
from auto_ml_kinder.classes.ModelTrainingData import ModelTrainingData
from bayes_opt import BayesianOptimization

class ModelTrainer():
    performance_df:pd.DataFrame = None
    data:ModelTrainingData = None
    models:list[ModelMeta] = []
    neural_network_bayesian_optimization:NeuralNetwork_BayesianOptimization = None
    func = None
    def __init__(self,data:ModelTrainingData) -> None:
        self.data = data
        self.neural_network_bayesian_optimization = NeuralNetwork_BayesianOptimization()
        self.func = Functions()

    def perform_operation_regression(self, permutate_n_less_column = 0, exclude_models: list[ModelAndParamRegression] = []):
        for model_and_param in ModelAndParamRegression:
            skip_this_model = any(exclude_model.name == model_and_param.name for exclude_model in exclude_models)
            if(skip_this_model):
                continue
            model,param = self.func.get_model_and_param_regression(power=ModelPower.HIGH,model_and_param=model_and_param)
            for X_train, X_val, X_test,selected_columns in self.data.generate_permutations_train(min_columns=len(self.data.X_original.columns)-permutate_n_less_column):
                best_param,best_model,score = self.func.train_test_random_search_regression(model=model,param_distributions=param,X_train=X_train,y_train=self.data.Y_train,X_test=X_val,y_test=self.data.Y_val)
                self.predictor_regression(model_and_param.name, best_param, best_model,X_test,selected_columns=selected_columns)

    def perform_operation_classification(self, permutate_n_less_column = 0, exclude_models: list[ModelAndParamClassifiction] = []):
        for model_and_param in ModelAndParamClassifiction:
            skip_this_model = any(exclude_model.name == model_and_param.name for exclude_model in exclude_models)
            if(skip_this_model):
                continue
            model,param = self.func.get_model_and_param_classification(power=ModelPower.HIGH,model_and_param=model_and_param)
            for X_train, X_val, X_test,selected_columns in self.data.generate_permutations_train(min_columns=len(self.data.X_original.columns)-permutate_n_less_column):
                best_param,best_model,score = self.func.train_test_random_search_classification(model=model,param_distributions=param,X_train=X_train,y_train=self.data.Y_train,X_test=X_val,y_test=self.data.Y_val)
                self.predictor_classification(model_and_param.name, best_param, best_model,X_test,selected_columns=selected_columns)

    def predictor_classification(self, model_name, best_param, best_model, X_test, selected_columns=[]):
        from sklearn.metrics import accuracy_score, f1_score
        import numpy as np
        try:
            total_columns = len(X_test[0])
        except:
            total_columns = len(X_test.columns)
        
        y_pred = best_model.predict(X_test)
        test_score = best_model.score(X_test, self.data.Y_test)
        accuracy = accuracy_score(self.data.Y_test, y_pred)
        f1 = f1_score(self.data.Y_test, y_pred, average='weighted')

        num_of_clusters = self.func.extract_last_digit_from_list(selected_columns, 'cluster')
        
        model_performance = ModelPerformance(
            score=test_score,
            model_name=model_name,
            accuracy=accuracy,
            f1_score=f1,
            total_columns=total_columns,
            scaler_type=type(self.data.Normalizer).__name__,
            selected_features=', '.join(selected_columns),
            num_of_clusters=num_of_clusters
        )
        
        self.models.append(ModelMeta(best_model, best_param))
        self.performance_df = self.func.insert_object_columns(self.performance_df, model_performance)

    def predictor_regression(self, model_name, best_param, best_model,X_test,selected_columns = []):
        from sklearn.metrics import mean_squared_error
        try:
            total_columns = len(X_test[0])
        except:
            total_columns = len(X_test.columns)
            
        y_pred = best_model.predict(X_test)
        test_score = best_model.score(X_test, self.data.Y_test)

        rmse = mean_squared_error(self.data.Y_test, y_pred, squared=False)
        num_of_clusters = self.func.extract_last_digit_from_list(selected_columns,'cluster')
        model_performance = ModelPerformance(score=test_score,model_name=model_name,RMSE=rmse,total_columns=total_columns,scaler_type=type(self.data.Normalizer).__name__,selected_features=', '.join(selected_columns),num_of_clusters=num_of_clusters)
        self.models.append(ModelMeta(best_model,best_param))
        self.performance_df = self.func.insert_object_columns(self.performance_df,model_performance)

    def predict_test_data(self,X,model):
        import numpy as np
        X_result = None

        if(self.data.create_clustering_feature_and_no_of_clusters[0]):
            labels = self.data.only_predict(X,self.data.clustering_model)
            if isinstance(X, pd.DataFrame):
                X_result = X.copy()
                X_result[f'cluster_{self.data.create_clustering_feature_and_no_of_clusters[1]}'] = labels
            elif isinstance(X, np.ndarray):
                X_result = np.hstack((X, labels.reshape(-1, 1)))
            elif isinstance(X, list):
                X_result = [row + [label] for row, label in zip(X, labels)]
        
        X_result = self.data.Data_transformer_pipe.transform(X_result)

        predictions = model.predict(X_result)

        return predictions

    def perform_neural_network_regression(self
                                          ,totalExperiments = 4
                                          ,params:NeuralNetwork_BayesianOptimization_Params = NeuralNetwork_BayesianOptimization_Params(
                                              neurons_min_max= (32, 128),
                                              batch_size_min_max=(32, 64),
                                              dropout_rate_min_max=(0.2,0.6),
                                              epochs_min_max=(50, 100),
                                              hidden_layers_min_max=(1,6),
                                              learning_rate_min_max=(0.001, .01),
                                              normalization_min_max=(0,1)
                                          )):
        PARAMS = {
            'neurons': params.neurons_min_max,
            'activation':params.activation_min_max,
            'optimizer':params.optimizer_min_max,
            'learning_rate':params.learning_rate_min_max,
            'batch_size':params.batch_size_min_max,
            'epochs':params.epochs_min_max,
            'normalization':params.normalization_min_max,
            'dropout':params.dropout_min_max,
            'dropout_rate':params.dropout_rate_min_max,
            'hidden_layers':params.hidden_layers_min_max
        }

        self.neural_network_bayesian_optimization = NeuralNetwork_BayesianOptimization()
        self.neural_network_bayesian_optimization.nn_regressor =  NeuralNetwork_Regression(self.data)
        
        # Run Bayesian Optimization
        self.neural_network_bayesian_optimization.nn_maximised = BayesianOptimization(self.neural_network_bayesian_optimization.nn_regressor.nn_cl_bo2, PARAMS, random_state=111,verbose=2)
        init,iter = int(totalExperiments / 2),int(totalExperiments / 2)
        self.neural_network_bayesian_optimization.nn_maximised.maximize(init_points=init, n_iter=iter)

    def perform_neural_network_classification(self
                                          ,totalExperiments = 4
                                          ,params:NeuralNetwork_BayesianOptimization_Params = NeuralNetwork_BayesianOptimization_Params(
                                              neurons_min_max= (32, 128),
                                              batch_size_min_max=(32, 64),
                                              dropout_rate_min_max=(0.2,0.6),
                                              epochs_min_max=(50, 100),
                                              hidden_layers_min_max=(1,6),
                                              learning_rate_min_max=(0.001, .01),
                                              normalization_min_max=(0,1)
                                          )):
        PARAMS = {
            'neurons': params.neurons_min_max,
            'activation':params.activation_min_max,
            'optimizer':params.optimizer_min_max,
            'learning_rate':params.learning_rate_min_max,
            'batch_size':params.batch_size_min_max,
            'epochs':params.epochs_min_max,
            'normalization':params.normalization_min_max,
            'dropout':params.dropout_min_max,
            'dropout_rate':params.dropout_rate_min_max,
            'hidden_layers':params.hidden_layers_min_max
        }

        self.neural_network_bayesian_optimization = NeuralNetwork_BayesianOptimization()
        self.neural_network_bayesian_optimization.nn_classifier =  NeuralNetwork_Classification(self.data)
        
        # Run Bayesian Optimization
        self.neural_network_bayesian_optimization.nn_maximised = BayesianOptimization(self.neural_network_bayesian_optimization.nn_classifier.nn_cl_bo2, PARAMS, random_state=111,verbose=2)
        init,iter = int(totalExperiments / 2),int(totalExperiments / 2)
        self.neural_network_bayesian_optimization.nn_maximised.maximize(init_points=init, n_iter=iter)


    def neural_network_best_model_regression(self,epochs = None):
        params_nn = self.neural_network_bayesian_optimization.nn_maximised.max['params']
        if epochs is not None:
            params_nn['epochs'] = epochs
        predictor = self.neural_network_bayesian_optimization.nn_regressor.get_best_model(params_nn)
        predictor.fit(self.data.X_train, self.data.Y_train)
        self.predictor_regression(model_name='NeuralNetwork',best_model=predictor,best_param=predictor.get_params(),X_test=self.data.X_test,selected_columns=self.data.Selected_Features)

    def neural_network_best_model_classification(self,epochs = None):
        import numpy as np
        import keras as k
        params_nn = self.neural_network_bayesian_optimization.nn_maximised.max['params']
        if epochs is not None:
            params_nn['epochs'] = epochs
        predictor = self.neural_network_bayesian_optimization.nn_classifier.get_best_model(params_nn)
        num_classes = len(np.unique(self.data.Y))
        if num_classes > 2:
            predictor.fit(self.data.X_train, self.data.Y_train_neural_network)
        else:
            predictor.fit(self.data.X_train, self.data.Y_train)
        self.predictor_neural_network_classification(model_name='NeuralNetwork',best_model=predictor,best_param=predictor.get_params(),X_test=self.data.X_test,selected_columns=self.data.Selected_Features)

    def predictor_neural_network_classification(self, model_name, best_param, best_model, X_test, selected_columns=[]):
        from sklearn.metrics import accuracy_score, f1_score
        import numpy as np
        try:
            total_columns = len(X_test[0])
        except:
            total_columns = len(X_test.columns)
        
        y_pred = best_model.predict(X_test)
        if y_pred.ndim == 1:# because we have only single neuron in last layer then prediction will be single value
            y_pred_int = y_pred
        else:# we have more than 1 neuron in last layer then prediction will be multiple then get max value
            y_pred_int = np.argmax(y_pred, axis=1).astype(np.float32)
        
        if y_pred.ndim == 1:# because we have only single neuron in last layer then prediction will be single value
            y_test_int = self.data.Y_test_neural_network
        else:# we have more than 1 neuron in last layer then prediction will be multiple then get max value
            y_test_int = np.argmax(self.data.Y_test_neural_network, axis=1).astype(np.float32)

        test_score = best_model.score(X_test, self.data.Y_test_neural_network)
        accuracy = accuracy_score(y_test_int, y_pred_int)
        f1 = f1_score(y_test_int, y_pred_int, average='weighted')

        num_of_clusters = self.func.extract_last_digit_from_list(selected_columns, 'cluster')
        
        model_performance = ModelPerformance(
            score=test_score,
            model_name=model_name,
            accuracy=accuracy,
            f1_score=f1,
            total_columns=total_columns,
            scaler_type=type(self.data.Normalizer).__name__,
            selected_features=', '.join(selected_columns),
            num_of_clusters=num_of_clusters
        )
        
        self.models.append(ModelMeta(best_model, best_param))
        self.performance_df = self.func.insert_object_columns(self.performance_df, model_performance)
  


from enum import Enum
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler, QuantileTransformer
from sklearn.preprocessing import PolynomialFeatures, StandardScaler
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline
import itertools
from itertools import combinations
import math
from auto_ml_kinder import model_feature_selector as mfs
from auto_ml_kinder.classes import ScalerType as st

def get_x_y(df:pd.DataFrame):
    X = df.iloc[:,0:-1]
    Y = df.iloc[:,-1]
    return X,Y
def get_X(df:pd.DataFrame):
    return df.iloc[:,0:-1]
def custom_train_val_test_df_split(df):
    X,Y = get_x_y(df)
    X_train,X_test,Y_train,Y_test = train_test_split(X,Y,test_size=.2,random_state=1)
    X_train,X_val,Y_train,Y_val = train_test_split(X_train,Y_train,test_size=.3,random_state=1)
    train_df = pd.concat([X_train,Y_train], axis=1)
    val_df = pd.concat([X_val,Y_val], axis=1)
    test_df = pd.concat([X_test,Y_test],axis=1)
    return train_df,val_df,test_df
def get_scaler(scaler_type):
    if not isinstance(scaler_type, st.ScalerType):
        raise ValueError("Invalid scaler type. Please provide a valid ScalerType enum value.")
    scaler = None
    if scaler_type == st.STANDARD_SCALER:
        scaler = StandardScaler()
    elif scaler_type == st.MINMAX_SCALER:
        scaler = MinMaxScaler()
    elif scaler_type == st.MAXABS_SCALER:
        scaler = MaxAbsScaler()
    elif scaler_type == st.ROBUST_SCALER:
        scaler = RobustScaler()
    elif scaler_type == st.QUANTILE_TRANSFORMER:
        scaler = QuantileTransformer()
    if scaler is None:
        raise ValueError("Invalid scaler type.")
    return scaler
    X_original:pd.DataFrame = None
    num_features = 0
    X:pd.DataFrame = None
    Y:pd.Series = None
    X_train_df:pd.DataFrame = None
    X_val_df:pd.DataFrame = None
    X_test_df:pd.DataFrame = None
    X_train:list = []
    X_val:list = []
    Y_train:list = []
    Y_train_neural_network:list = []
    Y_val_neural_network:list = []
    X_test:list = []
    Y_test_neural_network:list = []
    Normalizer = None
    Polynomializer = None
    PCAlizer = None
    Data_transformer_pipe:Pipeline = None
    Use_PCA = False
    Use_Polynomials = False
    Use_Feature_Selection = False
    create_clustering_feature_and_no_of_clusters = None
    clustering_model = None
    Selected_Features = []
    is_classification = False
    def __init__(self
                 ,df:pd.DataFrame
                 ,scaler_type:ScalerType
                 ,pca_variance = .95,
                 use_pca = False,
                 use_polynomials = False,
                 use_feature_selection = False,
                 create_clustering_feature_and_no_of_clusters = (False,3),
                 is_classification = False
                 ) -> None:
        self.Use_PCA = use_pca
        self.Use_Polynomials = use_polynomials
        self.Use_Feature_Selection = use_feature_selection
        self.create_clustering_feature_and_no_of_clusters = create_clustering_feature_and_no_of_clusters
        self.Selected_Features = df.columns
        self.is_classification = is_classification
        if(self.Use_Feature_Selection):
            tempX,tempY = get_x_y(df=df)
            columns_after_feature_selection = mfs.Run_Features_Selection(tempX,tempY)
            columns_after_feature_selection.append(tempY.name)# we need to add target column back to dataframe
            printable_names = ', '.join(columns_after_feature_selection)
            print(f'Total columns :{len(df.columns)}')
            print(f'Selected number of columns: {len(columns_after_feature_selection)}')
            print(f'Columns to use: {printable_names}')
            df = df[columns_after_feature_selection]
            self.Selected_Features = columns_after_feature_selection
            
        self.X,self.Y = get_x_y(df=df)
        self.X_original = self.X.copy(deep=True)
        self.num_features = self.X_original.shape[1]
        self.X_train_df,self.X_val_df,self.X_test_df = custom_train_val_test_df_split(df)
        self.X_train,self.Y_train = get_x_y(self.X_train_df)
        self.X_val,self.Y_val = get_x_y(self.X_val_df)
        self.X_test,self.Y_test = get_x_y(self.X_test_df)
        self.Y_train_neural_network = self.Y_train.copy()
        self.Y_test_neural_network = self.Y_test.copy()
        self.Y_val_neural_network = self.Y_val.copy()
        
        self.X_train_df,self.X_test_df,self.X_val_df = get_X(self.X_train_df),get_X(self.X_test_df),get_X(self.X_val_df)
        self.Normalizer = get_scaler(scaler_type)
        self.Polynomializer = PolynomialFeatures(degree=2)
        self.PCAlizer = PCA(n_components=pca_variance)
        total_columns = len(self.X_original.columns)
        pipeline = Pipeline([
            ('scaler', self.Normalizer)
        ])
        if(self.Use_Polynomials):
            pipeline.steps.append(('poly_features', self.Polynomializer))
        if(self.Use_PCA):
            pipeline.steps.append(('pca', self.PCAlizer))
            total_columns = pca_variance
        
        
        if(create_clustering_feature_and_no_of_clusters[0]):
            model = find_best_clustering_algorithm(self.X,num_clusters=create_clustering_feature_and_no_of_clusters[1])
            self.X,self.X_train,self.X_test,self.X_val,self.X_original,self.X_train_df,self.X_test_df,self.X_val_df = perform_clustering_on_first(
                self.X
                ,self.X_train
                ,self.X_test
                ,self.X_val
                ,self.X_original
                ,self.X_train_df
                ,self.X_test_df
                ,self.X_val_df
                ,num_clusters=create_clustering_feature_and_no_of_clusters[1]
                ,model=model)
            self.clustering_model = model
            self.Selected_Features = self.X_train_df.columns
            
        self.Data_transformer_pipe = pipeline
        
        if len(self.Data_transformer_pipe.steps) > 1:
            self.Data_transformer_pipe.fit(self.X_original)
            self.X = self.Data_transformer_pipe.transform(self.X)
            self.X_train = self.Data_transformer_pipe.transform(self.X_train)
            self.X_test = self.Data_transformer_pipe.transform(self.X_test)
            self.X_val = self.Data_transformer_pipe.transform(self.X_val)
            # self.X,self.X_train,self.X_test,self.X_val,self.X_original = perform_clustering_on_first(self.X,self.X_train,self.X_test,self.X_val,self.X_original)
        try:
            total_columns = len(self.X[0])
        except:
            total_columns = len(self.X.columns)
        print(f'Total columns being used after all data transformations: {total_columns}')
        if(self.is_classification):
            import keras as k
            num_classes = len(np.unique(self.Y))
            
            if num_classes > 2:
                print(f'Since num of classes is {num_classes} transforming Y_(test/train/val)_neural_network variables to categorical.')
                self.Y_test_neural_network = k.utils.to_categorical(self.Y_test)
                self.Y_train_neural_network = k.utils.to_categorical(self.Y_train)
                self.Y_val_neural_network = k.utils.to_categorical(self.Y_val)
    def generate_permutations_train(self, min_columns):
        num_features = self.X_original.shape[1]
        total_permutations = sum(len(list(combinations(range(num_features), r))) for r in range(min_columns, num_features + 1))
        print("Total Permutations:", total_permutations)
        columns = self.X_original.columns
        for r in range(min_columns, num_features + 1):
            for combo in combinations(range(num_features), r):
                selected_columns = [columns[i] for i in list(combo)]
                if len(self.Data_transformer_pipe.steps) == 0:
                    X_train_permuted = self.X_train_df.loc[:, selected_columns].values
                    X_val_permuted = self.X_val_df.loc[:, selected_columns].values
                    X_test_permuted = self.X_test_df.loc[:, selected_columns].values
                    # yield X_train_permuted, X_val_permuted, X_test_permuted
                else:
                    X_train_permuted = self.Data_transformer_pipe.fit_transform(self.X_train_df.loc[:, selected_columns])
                    X_val_permuted = self.Data_transformer_pipe.transform(self.X_val_df.loc[:, selected_columns])
                    X_test_permuted = self.Data_transformer_pipe.transform(self.X_test_df.loc[:, selected_columns])
                yield X_train_permuted, X_val_permuted, X_test_permuted,selected_columns
    def only_predict(self,X, clusterer):
        labels = clusterer.predict(X) if hasattr(clusterer, 'predict') else clusterer.fit_predict(X)
        return labels
import pandas as pd
import numpy as np
from sklearn.cluster import KMeans
def perform_clustering_on_first(*X_data, num_clusters=10,model):
    if len(X_data) == 0:
        raise ValueError("At least one dataset must be provided.")
    
    result_data = []
    # return X_data
    scores = list()
    for X in X_data:
        labels = fit_and_predict(X,model)
        if isinstance(X, pd.DataFrame):
            X_with_clusters = X.copy()
            X_with_clusters[f'cluster_{num_clusters}'] = labels
            result_data.append(X_with_clusters)
        elif isinstance(X, np.ndarray):
            X_with_clusters = np.hstack((X, labels.reshape(-1, 1)))
            result_data.append(X_with_clusters)
        elif isinstance(X, list):
            X_with_clusters = [row + [label] for row, label in zip(X, labels)]
            result_data.append(X_with_clusters)
        from sklearn.metrics import silhouette_score
        scores.append(silhouette_score(X, labels))
    mean_score = sum(scores) / len(scores)
    print(f'Mean score of all clusters is: {mean_score}')
    return tuple(result_data)
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans, DBSCAN, AgglomerativeClustering
from sklearn.metrics import silhouette_score
def find_best_clustering_algorithm(X, num_clusters=10):
    algorithms = {
        'kmeans': tune_kmeans(X,num_clusters),
        'agglomerative': tune_agglomerative(X,num_clusters),
        # 'spectral': tune_spectral(X),
        'mean_shift': tune_mean_shift(X),
        'dbscan': tune_dbscan(X),
        'birch': tune_birch(X),
        'affinity_propagation': tune_affinity_propagation(X),
        'mini_batch_kmeans': tune_mini_batch_kmeans(X,num_clusters)
    }
    best_score = -1
    best_algo_name = None
    best_algo = None
    
    for algo_name, clusterer in algorithms.items():
        labels = fit_and_predict(X, clusterer)
        num_labels = len(set(labels))
        if num_labels < 2 or num_labels >= len(X):
            print(f"Algorithm {algo_name} produced an invalid number of clusters: {num_labels}")
            continue
        score = silhouette_score(X, labels)
        print(f'Mean score for {algo_name} is: {score}')
        
        if score > best_score:
            best_score = score
            best_algo_name = algo_name
            best_algo = clusterer
    print(f'Best algorithm: {best_algo_name} with a score of: {best_score}')
    return algorithms[best_algo_name]
def fit_and_predict(X, clusterer):
    if hasattr(clusterer, 'predict'):
        clusterer.fit(X)
    else:
        clusterer.fit_predict(X)
    labels = clusterer.predict(X) if hasattr(clusterer, 'predict') else clusterer.fit_predict(X)
    return labels
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans, DBSCAN, AgglomerativeClustering
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import adjusted_rand_score, adjusted_mutual_info_score, completeness_score, homogeneity_score, v_measure_score
from sklearn.cluster import KMeans, AgglomerativeClustering, SpectralClustering, MeanShift, DBSCAN, Birch, AffinityPropagation, MiniBatchKMeans
def tune_kmeans(X, num_clusters=10):
    print("Grid tuning in progress for KMeans...")
    kmeans_params = {'n_clusters': [num_clusters], 'random_state': [42]}
    kmeans = KMeans()
    kmeans_gs = GridSearchCV(kmeans, kmeans_params, scoring=adjusted_rand_score)
    kmeans_gs.fit(X)
    return kmeans_gs.best_estimator_
def tune_agglomerative(X, num_clusters=10):
    print("Grid tuning in progress for Agglomerative Clustering...")
    agglomerative_params = {'n_clusters': [num_clusters], 'linkage': ['ward', 'complete', 'average', 'single']}
    agglomerative = AgglomerativeClustering()
    agglomerative_gs = GridSearchCV(agglomerative, agglomerative_params, scoring=adjusted_rand_score)
    agglomerative_gs.fit(X)
    return agglomerative_gs.best_estimator_
def tune_spectral(X, num_clusters_range=(3, 5)):
    print("Grid tuning in progress for Spectral Clustering...")
    spectral_params = {'n_clusters': list(range(*num_clusters_range))}
    spectral = SpectralClustering()
    spectral_gs = GridSearchCV(spectral, spectral_params, scoring=adjusted_rand_score)
    spectral_gs.fit(X)
    return spectral_gs.best_estimator_
def tune_mean_shift(X):
    # MeanShift does not require tuning hyperparameters
    print("Mean Shift clustering does not require hyperparameter tuning.")
    return MeanShift()
def tune_dbscan(X):
    # DBSCAN does not require tuning hyperparameters
    print("DBSCAN clustering does not require hyperparameter tuning.")
    return DBSCAN()
def tune_birch(X):
    # Birch does not require tuning hyperparameters
    print("Birch clustering does not require hyperparameter tuning.")
    return Birch()
def tune_affinity_propagation(X):
    # AffinityPropagation does not require tuning hyperparameters
    print("Affinity Propagation clustering does not require hyperparameter tuning.")
    return AffinityPropagation()
def tune_mini_batch_kmeans(X,num_clusters=3):
    print("Grid tuning in progress for Mini-Batch KMeans...")
    mini_batch_kmeans_params = {'n_clusters': [num_clusters], 'random_state': [42]}
    mini_batch_kmeans = MiniBatchKMeans()
    mini_batch_kmeans_gs = GridSearchCV(mini_batch_kmeans, mini_batch_kmeans_params, scoring=adjusted_rand_score)
    mini_batch_kmeans_gs.fit(X)
    return mini_batch_kmeans_gs.best_estimator_


from auto_ml_kinder import model_training_data_prep as dp
from sklearn.model_selection import RandomizedSearchCV
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from auto_ml_kinder import model_training_data_prep as dp
from auto_ml_kinder import model_training_helper as mth
from auto_ml_kinder.classes import NeuralNetwork_Regression as nnr
from auto_ml_kinder.classes import NeuralNetwork_Classification as nnc
from enum import Enum
import pandas as pd
from sklearn.metrics import mean_squared_error
import warnings
from bayes_opt import BayesianOptimization
from auto_ml_kinder import model_list_helper as mlh
from auto_ml_kinder.classes import ModelAndParamRegression as mpr
from auto_ml_kinder.classes import ModelAndParamClassifiction as mpc
# Filter out specific warning messages
warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=FutureWarning)
def get_model_and_param_regression(power:mlh.ModelPower,model_and_param:mpr.ModelAndParamRegression):
    if not isinstance(model_and_param, mpr.ModelAndParamRegression):
        raise ValueError("Invalid model_and_param type. Please provide a valid ModelAndParam enum value.")
    model = None
    param = None
    if model_and_param == mpr.ModelAndParamRegression.Linear_Regression:
        param,model = mlh.get_parameters_linear_reg()
    elif model_and_param == mpr.ModelAndParamRegression.Ridge_Regression:
        param,model = mlh.get_parameters_ridge_fit(power=power)
    elif model_and_param == mpr.ModelAndParamRegression.Lasso_Regression:
        param,model = mlh.get_parameters_lasso_fit(power=power)
    elif model_and_param == mpr.ModelAndParamRegression.ElasticNet_Regression:
        param,model = mlh.get_parameters_elasticnet_fit(power=power)
    elif model_and_param == mpr.ModelAndParamRegression.SVR_Regression:
        param,model = mlh.get_parameters_svr_fit(power=power)
    elif model_and_param == mpr.ModelAndParamRegression.DecisionTree_Regressor:
        param,model = mlh.get_parameters_decision_tree_fit_reg(power=power)
    elif model_and_param == mpr.ModelAndParamRegression.RandomForest_Regressor:
        param,model = mlh.get_parameters_random_forest_fit_reg(power=power)
    elif model_and_param == mpr.ModelAndParamRegression.GradientBoosting_Regressor:
        param,model = mlh.get_parameters_gradient_boosting_fit_reg(power=power)
    elif model_and_param == mpr.ModelAndParamRegression.KNeighbors_Regressor:
        param,model = mlh.get_parameters_knn_reg(power=power)
    return model,param
def get_model_and_param_classification(power:mlh.ModelPower,model_and_param:mpc.ModelAndParamClassifiction):
    if not isinstance(model_and_param, mpc.ModelAndParamClassifiction):
        raise ValueError("Invalid model_and_param type. Please provide a valid ModelAndParam enum value.")
    model = None
    param = None
    if model_and_param == mpc.ModelAndParamClassifiction.Logistic_Regression:
        param,model = mlh.get_parameters_logistic_reg()
    elif model_and_param == mpc.ModelAndParamClassifiction.Ridge_Classifiction:
        param,model = mlh.get_parameters_ridge_classifier(power=power)
    elif model_and_param == mpc.ModelAndParamClassifiction.SVC_Classification:
        param,model = mlh.get_parameters_svc_fit(power=power)
    elif model_and_param == mpc.ModelAndParamClassifiction.DecisionTree_Classifiction:
        param,model = mlh.get_parameters_decision_tree_classifier(power=power)
    elif model_and_param == mpc.ModelAndParamClassifiction.RandomForest_Classifiction:
        param,model = mlh.get_parameters_random_forest_classifier(power=power)
    elif model_and_param == mpc.ModelAndParamClassifiction.GradientBoosting_Classifiction:
        param,model = mlh.get_parameters_gradient_boosting_classifier(power=power)
    elif model_and_param == mpc.ModelAndParamClassifiction.KNeighbors_Classifiction:
        param,model = mlh.get_parameters_knn_classifier(power=power)
    return model,param
def train_test_random_search_regression(model, param_distributions, X_train, y_train, X_test, y_test, scoring='r2', cv=5):
    total_combinations = 1
    for values in param_distributions.values():
        total_combinations *= len(values)
    # Set n_iter based on the total_combinations
    n_iter = min(total_combinations, 10)
    random_search = RandomizedSearchCV(estimator=model, param_distributions=param_distributions,n_iter=n_iter, scoring=scoring, cv=cv, random_state=42)
    random_search.fit(X_train, y_train)
    
    best_params = random_search.best_params_
    best_model = random_search.best_estimator_
    
    best_model.fit(X_train, y_train)
    
    test_score = best_model.score(X_test, y_test)
    
    return best_params, best_model, test_score
def train_test_random_search_classification(model, param_distributions, X_train, y_train, X_test, y_test, scoring='accuracy', cv=5):
    total_combinations = 1
    for values in param_distributions.values():
        total_combinations *= len(values)
    
    # Set n_iter based on the total_combinations
    n_iter = min(total_combinations, 10)
    random_search = RandomizedSearchCV(estimator=model, param_distributions=param_distributions, n_iter=n_iter, scoring=scoring, cv=cv, random_state=42)
    random_search.fit(X_train, y_train)
    
    best_params = random_search.best_params_
    best_model = random_search.best_estimator_
    
    best_model.fit(X_train, y_train)
    
    test_score = best_model.score(X_test, y_test)
    
    return best_params, best_model, test_score
class ModelPerformance():
    score = None
    model_name = None
    RMSE = None
    total_columns = None
    scaler_type = None
    selected_features = None
    num_of_clusters_if_used = None
    accuracy = None
    f1_score = None
    def __init__(self,score,model_name,RMSE = None,total_columns = 0,scaler_type = None,selected_features = None,num_of_clusters = None, accuracy = None, f1_score = None) -> None:
        self.model_name = model_name
        self.score = score
        self.RMSE = RMSE
        self.total_columns = total_columns
        self.scaler_type = scaler_type
        self.selected_features = selected_features
        self.num_of_clusters_if_used = num_of_clusters
        self.accuracy = accuracy
        self.f1_score = f1_score
    def to_dict(self):
        return vars(self)
        
def insert_object_columns(df, obj:ModelPerformance):
    if df is None:
        df = pd.DataFrame(columns=obj.to_dict())
    
    row_data = obj.to_dict()
    
    df_new = pd.concat([df, pd.DataFrame([row_data])], ignore_index=True)
    
    return df_new
class ModelMeta():
    model = None
    model_params = None
    
    def __init__(self,model,params) -> None:
        self.model = model
        self.model_params = params
class NeuralNetwork_BayesianOptimization():
    nn_maximised:BayesianOptimization = None
    nn_regressor:nnr.NeuralNetwork_Regression = None
    nn_classifier:nnc.NeuralNetwork_Classification = None
class NeuralNetwork_BayesianOptimization_Params():
    neurons_min_max = None
    activation_min_max = None
    optimizer_min_max = None
    learning_rate_min_max = None
    batch_size_min_max = None
    epochs_min_max = None
    normalization_min_max = None
    dropout_rate_min_max = None
    dropout_min_max = None
    hidden_layers_min_max = None
    def __init__(self
                 ,neurons_min_max = (32, 128)
                 ,learning_rate_min_max = (0.001, .01)
                 ,batch_size_min_max = (32, 64)
                 ,epochs_min_max = (50, 100)
                 ,normalization_min_max = (0,1)
                 ,dropout_rate_min_max = (0.2,0.6)
                 ,hidden_layers_min_max = (1,2)
                 ,dropout_min_max = (0,1)
                 ,activation_min_max = (0, 9) ):
        self.neurons_min_max = neurons_min_max
        self.activation_min_max = activation_min_max
        self.optimizer_min_max = (0,6)
        self.learning_rate_min_max = learning_rate_min_max
        self.batch_size_min_max = batch_size_min_max
        self.epochs_min_max = epochs_min_max
        self.normalization_min_max = normalization_min_max
        self.dropout_min_max = dropout_min_max
        self.dropout_rate_min_max = dropout_rate_min_max
        self.hidden_layers_min_max = hidden_layers_min_max
        
class ModelTrainer():
    performance_df:pd.DataFrame = None
    data:dp.ModelTrainingData = None
    models:list[ModelMeta] = []
    neural_network_bayesian_optimization:NeuralNetwork_BayesianOptimization = None
    def __init__(self,data:dp.ModelTrainingData) -> None:
        self.data = data
        self.neural_network_bayesian_optimization = NeuralNetwork_BayesianOptimization()
    def perform_operation_regression(self, permutate_n_less_column = 0, exclude_models: list[mpr.ModelAndParamRegression] = []):
        for model_and_param in mpr.ModelAndParamRegression:
            skip_this_model = any(exclude_model.name == model_and_param.name for exclude_model in exclude_models)
            if(skip_this_model):
                continue
            model,param =get_model_and_param_regression(power=mlh.ModelPower.HIGH,model_and_param=model_and_param)
            for X_train, X_val, X_test,selected_columns in self.data.generate_permutations_train(min_columns=len(self.data.X_original.columns)-permutate_n_less_column):
                best_param,best_model,score = train_test_random_search_regression(model=model,param_distributions=param,X_train=X_train,y_train=self.data.Y_train,X_test=X_val,y_test=self.data.Y_val)
                self.predictor_regression(model_and_param.name, best_param, best_model,X_test,selected_columns=selected_columns)
    def perform_operation_classification(self, permutate_n_less_column = 0, exclude_models: list[mpc.ModelAndParamClassifiction] = []): 
        for model_and_param in mpc.ModelAndParamClassifiction:
            skip_this_model = any(exclude_model.name == model_and_param.name for exclude_model in exclude_models)
            if(skip_this_model):
                continue
            model,param = get_model_and_param_classification(power=mlh.ModelPower.HIGH,model_and_param=model_and_param)
            for X_train, X_val, X_test,selected_columns in self.data.generate_permutations_train(min_columns=len(self.data.X_original.columns)-permutate_n_less_column):
                best_param,best_model,score = train_test_random_search_classification(model=model,param_distributions=param,X_train=X_train,y_train=self.data.Y_train,X_test=X_val,y_test=self.data.Y_val)
                self.predictor_classification(model_and_param.name, best_param, best_model,X_test,selected_columns=selected_columns)
    def predictor_classification(self, model_name, best_param, best_model, X_test, selected_columns=[]):
        from sklearn.metrics import accuracy_score, f1_score
        import numpy as np
        try:
            total_columns = len(X_test[0])
        except:
            total_columns = len(X_test.columns)
        
        y_pred = best_model.predict(X_test)
        test_score = best_model.score(X_test, self.data.Y_test)
        accuracy = accuracy_score(self.data.Y_test, y_pred)
        f1 = f1_score(self.data.Y_test, y_pred, average='weighted')
        num_of_clusters = extract_last_digit_from_list(selected_columns, 'cluster')
        
        model_performance = ModelPerformance(
            score=test_score,
            model_name=model_name,
            accuracy=accuracy,
            f1_score=f1,
            total_columns=total_columns,
            scaler_type=type(self.data.Normalizer).__name__,
            selected_features=', '.join(selected_columns),
            num_of_clusters=num_of_clusters
        )
        
        self.models.append(ModelMeta(best_model, best_param))
        self.performance_df = insert_object_columns(self.performance_df, model_performance)
    def predictor_regression(self, model_name, best_param, best_model,X_test,selected_columns = []):
        from sklearn.metrics import mean_squared_error
        try:
            total_columns = len(X_test[0])
        except:
            total_columns = len(X_test.columns)
            
        y_pred = best_model.predict(X_test)
        test_score = best_model.score(X_test, self.data.Y_test)
        rmse = mean_squared_error(self.data.Y_test, y_pred, squared=False)
        num_of_clusters = extract_last_digit_from_list(selected_columns,'cluster')
        model_performance = ModelPerformance(score=test_score,model_name=model_name,RMSE=rmse,total_columns=total_columns,scaler_type=type(self.data.Normalizer).__name__,selected_features=', '.join(selected_columns),num_of_clusters=num_of_clusters)
        self.models.append(ModelMeta(best_model,best_param))
        self.performance_df = insert_object_columns(self.performance_df,model_performance)
    def predict_test_data(self,X,model):
        import numpy as np
        X_result = None
        if(self.data.create_clustering_feature_and_no_of_clusters[0]):
            labels = self.data.only_predict(X,self.data.clustering_model)
            if isinstance(X, pd.DataFrame):
                X_result = X.copy()
                X_result[f'cluster_{self.data.create_clustering_feature_and_no_of_clusters[1]}'] = labels
            elif isinstance(X, np.ndarray):
                X_result = np.hstack((X, labels.reshape(-1, 1)))
            elif isinstance(X, list):
                X_result = [row + [label] for row, label in zip(X, labels)]
        
        X_result = self.data.Data_transformer_pipe.transform(X_result)
        predictions = model.predict(X_result)
        return predictions
    def perform_neural_network_regression(self
                                          ,totalExperiments = 4
                                          ,params:NeuralNetwork_BayesianOptimization_Params = NeuralNetwork_BayesianOptimization_Params(
                                              neurons_min_max= (32, 128),
                                              batch_size_min_max=(32, 64),
                                              dropout_rate_min_max=(0.2,0.6),
                                              epochs_min_max=(50, 100),
                                              hidden_layers_min_max=(1,6),
                                              learning_rate_min_max=(0.001, .01),
                                              normalization_min_max=(0,1)
                                          )):
        PARAMS = {
            'neurons': params.neurons_min_max,
            'activation':params.activation_min_max,
            'optimizer':params.optimizer_min_max,
            'learning_rate':params.learning_rate_min_max,
            'batch_size':params.batch_size_min_max,
            'epochs':params.epochs_min_max,
            'normalization':params.normalization_min_max,
            'dropout':params.dropout_min_max,
            'dropout_rate':params.dropout_rate_min_max,
            'hidden_layers':params.hidden_layers_min_max
        }
        self.neural_network_bayesian_optimization = NeuralNetwork_BayesianOptimization()
        self.neural_network_bayesian_optimization.nn_regressor =  nnr.NeuralNetwork_Regression(self.data)
        
        # Run Bayesian Optimization
        self.neural_network_bayesian_optimization.nn_maximised = BayesianOptimization(f=self.neural_network_bayesian_optimization.nn_regressor.nn_cl_bo2, pbounds=PARAMS, random_state=111,verbose=2)
        init,iter = int(totalExperiments / 2),int(totalExperiments / 2)
        self.neural_network_bayesian_optimization.nn_maximised.maximize(init_points=init, n_iter=iter)
    def perform_neural_network_classification(self
                                          ,totalExperiments = 4
                                          ,params:NeuralNetwork_BayesianOptimization_Params = NeuralNetwork_BayesianOptimization_Params(
                                              neurons_min_max= (32.0, 128.0),
                                              batch_size_min_max=(32.0, 64.0),
                                              dropout_rate_min_max=(0.2,0.6),
                                              epochs_min_max=(50.0, 100.0),
                                              hidden_layers_min_max=(1.0,6.0),
                                              learning_rate_min_max=(0.001, .01),
                                              normalization_min_max=(0.0,1.0)
                                          )):
        PARAMS = {
            'neurons': params.neurons_min_max,
            'activation':params.activation_min_max,
            'optimizer':params.optimizer_min_max,
            'learning_rate':params.learning_rate_min_max,
            'batch_size':params.batch_size_min_max,
            'epochs':params.epochs_min_max,
            'normalization':params.normalization_min_max,
            'dropout':params.dropout_min_max,
            'dropout_rate':params.dropout_rate_min_max,
            'hidden_layers':params.hidden_layers_min_max
        }
        self.neural_network_bayesian_optimization = NeuralNetwork_BayesianOptimization()
        self.neural_network_bayesian_optimization.nn_classifier =  nnc.NeuralNetwork_Classification(self.data)
        
        # Run Bayesian Optimization
        self.neural_network_bayesian_optimization.nn_maximised = BayesianOptimization(f=self.neural_network_bayesian_optimization.nn_classifier.nn_cl_bo2, pbounds=PARAMS, random_state=111,verbose=2)
        init,iter = int(totalExperiments / 2),int(totalExperiments / 2)
        print('before maximise')
        self.neural_network_bayesian_optimization.nn_maximised.maximize(init_points=init, n_iter=iter)
        print('after maximise')
    def neural_network_best_model_regression(self,epochs = None):
        params_nn = self.neural_network_bayesian_optimization.nn_maximised.max['params']
        if epochs is not None:
            params_nn['epochs'] = epochs
        predictor = self.neural_network_bayesian_optimization.nn_regressor.get_best_model(params_nn)
        predictor.fit(self.data.X_train, self.data.Y_train)
        self.predictor_regression(model_name='NeuralNetwork',best_model=predictor,best_param=predictor.get_params(),X_test=self.data.X_test,selected_columns=self.data.Selected_Features)
    def neural_network_best_model_classification(self,epochs = None):
        import numpy as np
        import keras as k
        params_nn = self.neural_network_bayesian_optimization.nn_maximised.max['params']
        if epochs is not None:
            params_nn['epochs'] = epochs
        predictor = self.neural_network_bayesian_optimization.nn_classifier.get_best_model(params_nn)
        num_classes = len(np.unique(self.data.Y))
        if num_classes > 2:
            predictor.fit(self.data.X_train, self.data.Y_train_neural_network)
        else:
            predictor.fit(self.data.X_train, self.data.Y_train)
        self.predictor_neural_network_classification(model_name='NeuralNetwork',best_model=predictor,best_param=predictor.get_params(),X_test=self.data.X_test,selected_columns=self.data.Selected_Features)
    def predictor_neural_network_classification(self, model_name, best_param, best_model, X_test, selected_columns=[]):
        from sklearn.metrics import accuracy_score, f1_score
        import numpy as np
        try:
            total_columns = len(X_test[0])
        except:
            total_columns = len(X_test.columns)
        
        y_pred = best_model.predict(X_test)
        if y_pred.ndim == 1:# because we have only single neuron in last layer then prediction will be single value
            y_pred_int = y_pred
        else:# we have more than 1 neuron in last layer then prediction will be multiple then get max value
            y_pred_int = np.argmax(y_pred, axis=1).astype(np.float32)
        
        if y_pred.ndim == 1:# because we have only single neuron in last layer then prediction will be single value
            y_test_int = self.data.Y_test_neural_network
        else:# we have more than 1 neuron in last layer then prediction will be multiple then get max value
            y_test_int = np.argmax(self.data.Y_test_neural_network, axis=1).astype(np.float32)
        test_score = best_model.score(X_test, self.data.Y_test_neural_network)
        accuracy = accuracy_score(y_test_int, y_pred_int)
        f1 = f1_score(y_test_int, y_pred_int, average='weighted')
        num_of_clusters = extract_last_digit_from_list(selected_columns, 'cluster')
        
        model_performance = ModelPerformance(
            score=test_score,
            model_name=model_name,
            accuracy=accuracy,
            f1_score=f1,
            total_columns=total_columns,
            scaler_type=type(self.data.Normalizer).__name__,
            selected_features=', '.join(selected_columns),
            num_of_clusters=num_of_clusters
        )
        
        self.models.append(ModelMeta(best_model, best_param))
        self.performance_df = insert_object_columns(self.performance_df, model_performance)
    
def extract_last_digit_from_list(select_columns, match_string):
    # Initialize last_digit as None
    last_digit = None
    # Iterate through each column name in select_columns
    for column in select_columns:
        # Check if match_string is present in the column name
        if match_string in column:
            # Extract the last digit from the matched column name
            last_digit = column.split(match_string)[-1][-1]
            break  # Exit the loop after finding the first match
    return 0 if last_digit == None else int(last_digit)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.model_selection import train_test_split
from sklearn.model_selection import cross_val_score, KFold
from keras.api.models import Sequential
from keras.api.layers import Dense, BatchNormalization, Dropout
from keras.api.optimizers import Adam, SGD, RMSprop, Adadelta, Adagrad, Adamax, Nadam, Ftrl
from keras.api.callbacks import EarlyStopping, ModelCheckpoint
from math import floor
from sklearn.metrics import make_scorer, accuracy_score
from bayes_opt import BayesianOptimization
from sklearn.model_selection import StratifiedKFold
from keras.api.layers import LeakyReLU
LeakyReLU = LeakyReLU(alpha=0.1)
import warnings
warnings.filterwarnings('ignore')
pd.set_option("display.max_columns", None)
from scikeras.wrappers import KerasRegressor, KerasClassifier
from sklearn.metrics import mean_squared_error
from keras.api.callbacks import LearningRateScheduler
from auto_ml_kinder.classes.ModelTrainingData import ModelTrainingData


class NeuralNetwork_Classification():
    data:ModelTrainingData
    model_params = None

    def __init__(self,data:ModelTrainingData) -> None:
        self.data = data

    def get_nn_model(self,neurons,total_columns,activation,normalization,hidden_layers,dropout,dropout_rate,optimizer):
        nn = Sequential()
        nn.add(Dense(neurons, input_dim=total_columns, activation=activation))

        if normalization > 0.5:
            nn.add(BatchNormalization())

        for _ in range(hidden_layers):
            hidden_neurons = round(neurons / (_ + 1))
            nn.add(Dense(hidden_neurons, activation=activation))

            if normalization > 0.5:
                nn.add(BatchNormalization())

            if dropout == 1:
                nn.add(Dropout(rate=dropout_rate))
        
        num_classes = len(np.unique(self.data.Y))
    
        if num_classes == 2:
            # print(f'Total number of classes: {num_classes} using sigmoid as activation')
            nn.add(Dense(1, activation='sigmoid'))
            nn.compile(optimizer=optimizer, loss='binary_crossentropy', metrics=['accuracy'])
        else:
            # print(f'Total number of classes: {num_classes} using softmax as activation')
            nn.add(Dense(num_classes, activation='softmax'))
            nn.compile(optimizer=optimizer, loss='categorical_crossentropy', metrics=['accuracy'])

        return nn
    
    def lr_schedule(epoch, initial_lr):
        drop = 0.5
        epochs_drop = 10
        lr = initial_lr * (drop ** np.floor(epoch / epochs_drop))
        return lr

    # Create a LearningRateScheduler callback that uses the initial learning rate
    def create_lr_scheduler(self,initial_lr):
        return LearningRateScheduler(lambda epoch: self.lr_schedule(epoch, initial_lr))

    def nn_cl_bo2(self, neurons, activation, optimizer, learning_rate, batch_size, epochs, normalization, dropout, dropout_rate, hidden_layers):
        optimizerL = [
            'Adam', 'RMSprop', 'Adadelta', 'Adagrad', 'Adamax', 'Nadam', 'Ftrl'
        ]
        activationL = ['relu', 'selu', 'elu', 'sigmoid', 'softplus', 'softsign', 'tanh', 'exponential']
        
        neurons = round(neurons)
        activation = activationL[round(activation)]
        optimizer = optimizerL[round(optimizer)].lower()
        batch_size = round(batch_size)
        epochs = round(epochs)
        hidden_layers = round(hidden_layers)

        # Get total columns (features)
        try:
            total_columns = len(self.data.X_train[0])
        except:
            total_columns = len(self.data.X_train.columns)

        # Define the model building function
        def get_build_function():
            return self.get_nn_model(neurons, total_columns, activation, normalization, hidden_layers, dropout, dropout_rate, optimizer)

        # Early stopping callback (can be used if needed)
        es = EarlyStopping(monitor='accuracy', mode='max', verbose=0, patience=20)

        # Build the model
        nn = KerasClassifier(build_fn=get_build_function, epochs=epochs, batch_size=batch_size, verbose=0)

        # Split the dataset into training and testing sets (80% train, 20% test)
        X_train, X_test, y_train, y_test = train_test_split(self.data.X_train, self.data.Y_train_neural_network, test_size=0.2, random_state=42)

        # Train the model
        nn.fit(X_train, y_train, epochs=epochs, batch_size=batch_size, verbose=0, callbacks=[es])

        # Evaluate the model on the test set
        y_pred = nn.predict(X_test)
        accuracy = accuracy_score(y_test, y_pred)

        return accuracy

    def get_best_model(self,params):
        params_nn_ = params
        learning_rate = params_nn_['learning_rate']
        activationL = ['relu',LeakyReLU, 'selu','elu','sigmoid','softplus','softsign','tanh','exponential']
        activation = activationL[round(params_nn_['activation'])]
        batch_size = round(params_nn_['batch_size'])
        epochs = round(params_nn_['epochs'])
        normalization = round(params_nn_['normalization'])
        dropout = round(params_nn_['dropout'])
        dropout_rate = params_nn_['dropout_rate']
        neurons = round(params_nn_['neurons'])
        optimizerL = ['Adam', 'RMSprop', 'Adadelta', 'Adagrad', 'Adamax', 'Nadam', 'Ftrl','Adam']
        optimizer = optimizerL[round(params_nn_['optimizer'])].lower()
        hidden_layers = round(params_nn_['hidden_layers'])
        dropout = round(params_nn_['dropout'])
        
        try:
            total_columns = len(self.data.X_train[0])
        except:
            total_columns = len(self.data.X_train.columns)
        self.print_things(activation, batch_size, epochs, hidden_layers, normalization, dropout, dropout_rate, neurons, optimizer, total_columns)

        def get_build_function():
            return self.get_nn_model(neurons,total_columns,activation,normalization,hidden_layers,dropout,dropout_rate,optimizer)
        
        lr_scheduler = self.create_lr_scheduler(learning_rate)
        es = EarlyStopping(monitor='accuracy', mode='max', verbose=0, patience=20)
        nn_keras_classifier = KerasClassifier(build_fn=get_build_function, epochs=epochs, batch_size=batch_size,verbose=0)

        num_classes = len(np.unique(self.data.Y))
        if(num_classes > 1):
            nn_keras_classifier.fit(self.data.X_train, self.data.Y_train_neural_network, validation_data=(self.data.X_val, self.data.Y_val_neural_network), verbose=1, callbacks=[es,lr_scheduler])
        else:
            nn_keras_classifier.fit(self.data.X_train, self.data.Y_train, validation_data=(self.data.X_val, self.data.Y_val), verbose=1, callbacks=[es,lr_scheduler])

        return nn_keras_classifier

    def print_things(self, activation, batch_size, epochs, hidden_layers, normalization, dropout, dropout_rate, neurons, optimizer, total_columns,seperator = '\n'):
        stringToPrint = f'Neurons: {neurons}{seperator}Total Columns:{total_columns}{seperator}Normalization:{normalization > .5}{seperator}Activation:{activation}{seperator}Total hidden layers:{hidden_layers}{seperator}Dropout used:{dropout == 1}{seperator}Dropout rate if used:{dropout_rate}{seperator}Optimizer:{optimizer}{seperator}Epochs:{epochs}{seperator}Batch_size:{batch_size}\n'
        print(stringToPrint)



from enum import Enum
from sklearn.preprocessing import StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler, QuantileTransformer
from sklearn.preprocessing import StandardScaler

class ScalerType(Enum):
    STANDARD_SCALER = StandardScaler
    MINMAX_SCALER = MinMaxScaler
    MAXABS_SCALER = MaxAbsScaler
    ROBUST_SCALER = RobustScaler
    QUANTILE_TRANSFORMER = QuantileTransformer


import pandas as pd
from auto_ml_kinder.classes.ModelAndParamRegression import ModelAndParamRegression
from auto_ml_kinder.classes.ModelPower import ModelPower
from auto_ml_kinder.classes.ModelAndParamClassifiction import ModelAndParamClassifiction
from auto_ml_kinder.classes.ModelPerformance import ModelPerformance
from auto_ml_kinder.classes.ScalerType import ScalerType
from auto_ml_kinder.model_list_helper import get_parameters_decision_tree_classifier,get_parameters_decision_tree_fit_reg,get_parameters_elasticnet_fit,get_parameters_gradient_boosting_classifier,get_parameters_gradient_boosting_fit_reg,get_parameters_knn_classifier,get_parameters_knn_reg,get_parameters_lasso_fit,get_parameters_linear_reg,get_parameters_logistic_reg,get_parameters_random_forest_classifier,get_parameters_random_forest_fit_reg,get_parameters_ridge_classifier,get_parameters_ridge_fit,get_parameters_svc_fit,get_parameters_svr_fit
from sklearn.model_selection import RandomizedSearchCV
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler, QuantileTransformer
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans, DBSCAN, AgglomerativeClustering
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import adjusted_rand_score, adjusted_mutual_info_score, completeness_score, homogeneity_score, v_measure_score
from sklearn.cluster import KMeans, AgglomerativeClustering, SpectralClustering, MeanShift, DBSCAN, Birch, AffinityPropagation, MiniBatchKMeans
from sklearn.metrics import silhouette_score



class Functions():
    def get_model_and_param_regression(self,power:ModelPower,model_and_param:ModelAndParamRegression):
        if not isinstance(model_and_param, ModelAndParamRegression):
            raise ValueError("Invalid model_and_param type. Please provide a valid ModelAndParam enum value.")

        model = None
        param = None
        if model_and_param == ModelAndParamRegression.Linear_Regression:
            param,model = get_parameters_linear_reg()
        elif model_and_param == ModelAndParamRegression.Ridge_Regression:
            param,model = get_parameters_ridge_fit(power=power)
        elif model_and_param == ModelAndParamRegression.Lasso_Regression:
            param,model = get_parameters_lasso_fit(power=power)
        elif model_and_param == ModelAndParamRegression.ElasticNet_Regression:
            param,model = get_parameters_elasticnet_fit(power=power)
        elif model_and_param == ModelAndParamRegression.SVR_Regression:
            param,model = get_parameters_svr_fit(power=power)
        elif model_and_param == ModelAndParamRegression.DecisionTree_Regressor:
            param,model = get_parameters_decision_tree_fit_reg(power=power)
        elif model_and_param == ModelAndParamRegression.RandomForest_Regressor:
            param,model = get_parameters_random_forest_fit_reg(power=power)
        elif model_and_param == ModelAndParamRegression.GradientBoosting_Regressor:
            param,model = get_parameters_gradient_boosting_fit_reg(power=power)
        elif model_and_param == ModelAndParamRegression.KNeighbors_Regressor:
            param,model = get_parameters_knn_reg(power=power)

        return model,param

    def get_model_and_param_classification(self,power:ModelPower,model_and_param:ModelAndParamClassifiction):
        if not isinstance(model_and_param, ModelAndParamClassifiction):
            raise ValueError("Invalid model_and_param type. Please provide a valid ModelAndParam enum value.")

        model = None
        param = None
        if model_and_param == ModelAndParamClassifiction.Logistic_Regression:
            param,model = get_parameters_logistic_reg()
        elif model_and_param == ModelAndParamClassifiction.Ridge_Classifiction:
            param,model = get_parameters_ridge_classifier(power=power)
        elif model_and_param == ModelAndParamClassifiction.SVC_Classification:
            param,model = get_parameters_svc_fit(power=power)
        elif model_and_param == ModelAndParamClassifiction.DecisionTree_Classifiction:
            param,model = get_parameters_decision_tree_classifier(power=power)
        elif model_and_param == ModelAndParamClassifiction.RandomForest_Classifiction:
            param,model = get_parameters_random_forest_classifier(power=power)
        elif model_and_param == ModelAndParamClassifiction.GradientBoosting_Classifiction:
            param,model = get_parameters_gradient_boosting_classifier(power=power)
        elif model_and_param == ModelAndParamClassifiction.KNeighbors_Classifiction:
            param,model = get_parameters_knn_classifier(power=power)

        return model,param

    def train_test_random_search_regression(self,model, param_distributions, X_train, y_train, X_test, y_test, scoring='r2', cv=5):
        total_combinations = 1
        for values in param_distributions.values():
            total_combinations *= len(values)
        # Set n_iter based on the total_combinations
        n_iter = min(total_combinations, 10)

        random_search = RandomizedSearchCV(estimator=model, param_distributions=param_distributions,n_iter=n_iter, scoring=scoring, cv=cv, random_state=42)
        random_search.fit(X_train, y_train)
        
        best_params = random_search.best_params_
        best_model = random_search.best_estimator_
        
        best_model.fit(X_train, y_train)
        
        test_score = best_model.score(X_test, y_test)
        
        return best_params, best_model, test_score

    def train_test_random_search_classification(self,model, param_distributions, X_train, y_train, X_test, y_test, scoring='accuracy', cv=5):
        total_combinations = 1
        for values in param_distributions.values():
            total_combinations *= len(values)
        
        # Set n_iter based on the total_combinations
        n_iter = min(total_combinations, 10)

        random_search = RandomizedSearchCV(estimator=model, param_distributions=param_distributions, n_iter=n_iter, scoring=scoring, cv=cv, random_state=42)
        random_search.fit(X_train, y_train)
        
        best_params = random_search.best_params_
        best_model = random_search.best_estimator_
        
        best_model.fit(X_train, y_train)
        
        test_score = best_model.score(X_test, y_test)
        
        return best_params, best_model, test_score

        
    def insert_object_columns(self,df, obj:ModelPerformance):
        if df is None:
            df = pd.DataFrame(columns=obj.to_dict())
        
        row_data = obj.to_dict()
        
        df_new = pd.concat([df, pd.DataFrame([row_data])], ignore_index=True)
        
        return df_new

    
    def extract_last_digit_from_list(self,select_columns, match_string):
        # Initialize last_digit as None
        last_digit = None

        # Iterate through each column name in select_columns
        for column in select_columns:
            # Check if match_string is present in the column name
            if match_string in column:
                # Extract the last digit from the matched column name
                last_digit = column.split(match_string)[-1][-1]
                break  # Exit the loop after finding the first match

        return 0 if last_digit == None else int(last_digit)
    

    def get_x_y(self,df:pd.DataFrame):
        X = df.iloc[:,0:-1]
        Y = df.iloc[:,-1]
        return X,Y

    def get_X(self,df:pd.DataFrame):
        return df.iloc[:,0:-1]

    def custom_train_val_test_df_split(self,df):
        X,Y = self.get_x_y(df)
        X_train,X_test,Y_train,Y_test = train_test_split(X,Y,test_size=.2,random_state=1)
        X_train,X_val,Y_train,Y_val = train_test_split(X_train,Y_train,test_size=.3,random_state=1)
        train_df = pd.concat([X_train,Y_train], axis=1)
        val_df = pd.concat([X_val,Y_val], axis=1)
        test_df = pd.concat([X_test,Y_test],axis=1)
        return train_df,val_df,test_df



    def get_scaler(self,scaler_type):
        if not isinstance(scaler_type, ScalerType):
            raise ValueError("Invalid scaler type. Please provide a valid ScalerType enum value.")

        scaler = None
        if scaler_type == ScalerType.STANDARD_SCALER:
            scaler = StandardScaler()
        elif scaler_type == ScalerType.MINMAX_SCALER:
            scaler = MinMaxScaler()
        elif scaler_type == ScalerType.MAXABS_SCALER:
            scaler = MaxAbsScaler()
        elif scaler_type == ScalerType.ROBUST_SCALER:
            scaler = RobustScaler()
        elif scaler_type == ScalerType.QUANTILE_TRANSFORMER:
            scaler = QuantileTransformer()

        if scaler is None:
            raise ValueError("Invalid scaler type.")

        return scaler
    
    def find_best_clustering_algorithm(self,X, num_clusters=10):
        algorithms = {
            'kmeans': self.tune_kmeans(X,num_clusters),
            'agglomerative': self.tune_agglomerative(X,num_clusters),
            # 'spectral': tune_spectral(X),
            'mean_shift': self.tune_mean_shift(X),
            'dbscan': self.tune_dbscan(X),
            'birch': self.tune_birch(X),
            'affinity_propagation': self.tune_affinity_propagation(X),
            'mini_batch_kmeans': self.tune_mini_batch_kmeans(X,num_clusters)
        }
    def tune_kmeans(self,X, num_clusters=10):
        print("Grid tuning in progress for KMeans...")
        kmeans_params = {'n_clusters': [num_clusters], 'random_state': [42]}
        kmeans = KMeans()
        kmeans_gs = GridSearchCV(kmeans, kmeans_params, scoring=adjusted_rand_score)
        kmeans_gs.fit(X)
        return kmeans_gs.best_estimator_

    def tune_agglomerative(self,X, num_clusters=10):
        print("Grid tuning in progress for Agglomerative Clustering...")
        agglomerative_params = {'n_clusters': [num_clusters], 'linkage': ['ward', 'complete', 'average', 'single']}
        agglomerative = AgglomerativeClustering()
        agglomerative_gs = GridSearchCV(agglomerative, agglomerative_params, scoring=adjusted_rand_score)
        agglomerative_gs.fit(X)
        return agglomerative_gs.best_estimator_

    def tune_spectral(self,X, num_clusters_range=(3, 5)):
        print("Grid tuning in progress for Spectral Clustering...")
        spectral_params = {'n_clusters': list(range(*num_clusters_range))}
        spectral = SpectralClustering()
        spectral_gs = GridSearchCV(spectral, spectral_params, scoring=adjusted_rand_score)
        spectral_gs.fit(X)
        return spectral_gs.best_estimator_

    def tune_mean_shift(self,X):
        # MeanShift does not require tuning hyperparameters
        print("Mean Shift clustering does not require hyperparameter tuning.")
        return MeanShift()

    def tune_dbscan(self,X):
        # DBSCAN does not require tuning hyperparameters
        print("DBSCAN clustering does not require hyperparameter tuning.")
        return DBSCAN()

    def tune_birch(self,X):
        # Birch does not require tuning hyperparameters
        print("Birch clustering does not require hyperparameter tuning.")
        return Birch()

    def tune_affinity_propagation(self,X):
        # AffinityPropagation does not require tuning hyperparameters
        print("Affinity Propagation clustering does not require hyperparameter tuning.")
        return AffinityPropagation()

    def tune_mini_batch_kmeans(self,X,num_clusters=3):
        print("Grid tuning in progress for Mini-Batch KMeans...")
        mini_batch_kmeans_params = {'n_clusters': [num_clusters], 'random_state': [42]}
        mini_batch_kmeans = MiniBatchKMeans()
        mini_batch_kmeans_gs = GridSearchCV(mini_batch_kmeans, mini_batch_kmeans_params, scoring=adjusted_rand_score)
        mini_batch_kmeans_gs.fit(X)
        return mini_batch_kmeans_gs.best_estimator_

    def find_best_clustering_algorithm(self,X, num_clusters=10):
        algorithms = {
            'kmeans': self.tune_kmeans(X,num_clusters),
            'agglomerative': self.tune_agglomerative(X,num_clusters),
            # 'spectral': tune_spectral(X),
            'mean_shift': self.tune_mean_shift(X),
            'dbscan':self. tune_dbscan(X),
            'birch': self.tune_birch(X),
            'affinity_propagation': self.tune_affinity_propagation(X),
            'mini_batch_kmeans': self.tune_mini_batch_kmeans(X,num_clusters)
        }

        best_score = -1
        best_algo_name = None
        best_algo = None
        
        for algo_name, clusterer in algorithms.items():
            labels = self.fit_and_predict(X, clusterer)

            num_labels = len(set(labels))
            if num_labels < 2 or num_labels >= len(X):
                print(f"Algorithm {algo_name} produced an invalid number of clusters: {num_labels}")
                continue

            score = silhouette_score(X, labels)
            print(f'Mean score for {algo_name} is: {score}')
            
            if score > best_score:
                best_score = score
                best_algo_name = algo_name
                best_algo = clusterer

        print(f'Best algorithm: {best_algo_name} with a score of: {best_score}')
        return algorithms[best_algo_name]
    
    def fit_and_predict(self,X, clusterer):
        if hasattr(clusterer, 'predict'):
            clusterer.fit(X)
        else:
            clusterer.fit_predict(X)
        labels = clusterer.predict(X) if hasattr(clusterer, 'predict') else clusterer.fit_predict(X)
        return labels

    def perform_clustering_on_first(self,*X_data, num_clusters=10,model):
        if len(X_data) == 0:
            raise ValueError("At least one dataset must be provided.")
        
        result_data = []
        # return X_data
        scores = list()
        for X in X_data:
            labels = self.fit_and_predict(X,model)
            if isinstance(X, pd.DataFrame):
                X_with_clusters = X.copy()
                X_with_clusters[f'cluster_{num_clusters}'] = labels
                result_data.append(X_with_clusters)
            elif isinstance(X, np.ndarray):
                X_with_clusters = np.hstack((X, labels.reshape(-1, 1)))
                result_data.append(X_with_clusters)
            elif isinstance(X, list):
                X_with_clusters = [row + [label] for row, label in zip(X, labels)]
                result_data.append(X_with_clusters)

            from sklearn.metrics import silhouette_score
            scores.append(silhouette_score(X, labels))

        mean_score = sum(scores) / len(scores)
        print(f'Mean score of all clusters is: {mean_score}')
        return tuple(result_data)



from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression, RidgeClassifier
from sklearn.neighbors import KNeighborsClassifier
from enum import Enum

class ModelAndParamClassifiction(Enum):
    Logistic_Regression = LogisticRegression
    Ridge_Classifiction = RidgeClassifier
    SVC_Classification = SVC
    DecisionTree_Classifiction = DecisionTreeClassifier
    RandomForest_Classifiction = RandomForestClassifier
    GradientBoosting_Classifiction = GradientBoostingClassifier
    KNeighbors_Classifiction = KNeighborsClassifier


from bayes_opt import BayesianOptimization
from auto_ml_kinder.classes.NeuralNetwork_Regression import NeuralNetwork_Regression
from auto_ml_kinder.classes.NeuralNetwork_Classification import NeuralNetwork_Classification

class NeuralNetwork_BayesianOptimization():
    nn_maximised:BayesianOptimization = None
    nn_regressor:NeuralNetwork_Regression = None
    nn_classifier:NeuralNetwork_Classification = None


import pandas as pd
import pandas as pd
from sklearn.preprocessing import PolynomialFeatures
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline
from itertools import combinations
from auto_ml_kinder import model_feature_selector as mfs
from auto_ml_kinder.classes.Functions import Functions 
from auto_ml_kinder.classes.ScalerType import ScalerType
import numpy as np


class ModelTrainingData():
    X_original:pd.DataFrame = None
    num_features = 0
    X:pd.DataFrame = None
    Y:pd.Series = None
    X_train_df:pd.DataFrame = None
    X_val_df:pd.DataFrame = None
    X_test_df:pd.DataFrame = None

    X_train:list = []
    X_val:list = []
    Y_train:list = []
    Y_train_neural_network:list = []
    Y_val_neural_network:list = []

    X_test:list = []
    Y_test_neural_network:list = []

    Normalizer = None
    Polynomializer = None
    PCAlizer = None
    Data_transformer_pipe:Pipeline = None

    Use_PCA = False
    Use_Polynomials = False
    Use_Feature_Selection = False
    create_clustering_feature_and_no_of_clusters = None
    clustering_model = None
    Selected_Features = []
    is_classification = False
    def __init__(self
                 ,df:pd.DataFrame
                 ,scaler_type:ScalerType
                 ,pca_variance = .95,
                 use_pca = False,
                 use_polynomials = False,
                 use_feature_selection = False,
                 create_clustering_feature_and_no_of_clusters = (False,3),
                 is_classification = False
                 ) -> None:
        
        func = Functions()
        self.Use_PCA = use_pca
        self.Use_Polynomials = use_polynomials
        self.Use_Feature_Selection = use_feature_selection
        self.create_clustering_feature_and_no_of_clusters = create_clustering_feature_and_no_of_clusters
        self.Selected_Features = df.columns
        self.is_classification = is_classification
        if(self.Use_Feature_Selection):
            tempX,tempY = func.get_x_y(df=df)
            columns_after_feature_selection = mfs.Run_Features_Selection(tempX,tempY)
            columns_after_feature_selection.append(tempY.name)# we need to add target column back to dataframe
            printable_names = ', '.join(columns_after_feature_selection)
            print(f'Total columns :{len(df.columns)}')
            print(f'Selected number of columns: {len(columns_after_feature_selection)}')
            print(f'Columns to use: {printable_names}')
            df = df[columns_after_feature_selection]
            self.Selected_Features = columns_after_feature_selection
            
        self.X,self.Y = func.get_x_y(df=df)
        self.X_original = self.X.copy(deep=True)
        self.num_features = self.X_original.shape[1]
        self.X_train_df,self.X_val_df,self.X_test_df = func.custom_train_val_test_df_split(df)
        self.X_train,self.Y_train = func.get_x_y(self.X_train_df)
        self.X_val,self.Y_val = func.get_x_y(self.X_val_df)
        self.X_test,self.Y_test = func.get_x_y(self.X_test_df)

        self.Y_train_neural_network = self.Y_train.copy()
        self.Y_test_neural_network = self.Y_test.copy()
        self.Y_val_neural_network = self.Y_val.copy()
        
        self.X_train_df,self.X_test_df,self.X_val_df = func.get_X(self.X_train_df),func.get_X(self.X_test_df),func.get_X(self.X_val_df)

        self.Normalizer = func.get_scaler(scaler_type)
        self.Polynomializer = PolynomialFeatures(degree=2)
        self.PCAlizer = PCA(n_components=pca_variance)

        total_columns = len(self.X_original.columns)
        pipeline = Pipeline([
            ('scaler', self.Normalizer)
        ])
        if(self.Use_Polynomials):
            pipeline.steps.append(('poly_features', self.Polynomializer))
        if(self.Use_PCA):
            pipeline.steps.append(('pca', self.PCAlizer))
            total_columns = pca_variance
        
        
        if(create_clustering_feature_and_no_of_clusters[0]):
            model = func.find_best_clustering_algorithm(self.X,num_clusters=create_clustering_feature_and_no_of_clusters[1])
            self.X,self.X_train,self.X_test,self.X_val,self.X_original,self.X_train_df,self.X_test_df,self.X_val_df = func.perform_clustering_on_first(
                self.X
                ,self.X_train
                ,self.X_test
                ,self.X_val
                ,self.X_original
                ,self.X_train_df
                ,self.X_test_df
                ,self.X_val_df
                ,num_clusters=create_clustering_feature_and_no_of_clusters[1]
                ,model=model)
            self.clustering_model = model
            self.Selected_Features = self.X_train_df.columns
            
        self.Data_transformer_pipe = pipeline
        
        if len(self.Data_transformer_pipe.steps) > 1:
            self.Data_transformer_pipe.fit(self.X_original)
            self.X = self.Data_transformer_pipe.transform(self.X)
            self.X_train = self.Data_transformer_pipe.transform(self.X_train)
            self.X_test = self.Data_transformer_pipe.transform(self.X_test)
            self.X_val = self.Data_transformer_pipe.transform(self.X_val)

            # self.X,self.X_train,self.X_test,self.X_val,self.X_original = perform_clustering_on_first(self.X,self.X_train,self.X_test,self.X_val,self.X_original)

        try:
            total_columns = len(self.X[0])
        except:
            total_columns = len(self.X.columns)

        print(f'Total columns being used after all data transformations: {total_columns}')

        if(self.is_classification):
            import keras as k
            num_classes = len(np.unique(self.Y))
            
            if num_classes > 2:
                print(f'Since num of classes is {num_classes} transforming Y_(test/train/val)_neural_network variables to categorical.')
                self.Y_test_neural_network = k.utils.to_categorical(self.Y_test)
                self.Y_train_neural_network = k.utils.to_categorical(self.Y_train)
                self.Y_val_neural_network = k.utils.to_categorical(self.Y_val)

    def generate_permutations_train(self, min_columns):
        num_features = self.X_original.shape[1]
        total_permutations = sum(len(list(combinations(range(num_features), r))) for r in range(min_columns, num_features + 1))
        print("Total Permutations:", total_permutations)

        columns = self.X_original.columns
        for r in range(min_columns, num_features + 1):
            for combo in combinations(range(num_features), r):
                selected_columns = [columns[i] for i in list(combo)]
                if len(self.Data_transformer_pipe.steps) == 0:
                    X_train_permuted = self.X_train_df.loc[:, selected_columns].values
                    X_val_permuted = self.X_val_df.loc[:, selected_columns].values
                    X_test_permuted = self.X_test_df.loc[:, selected_columns].values
                    # yield X_train_permuted, X_val_permuted, X_test_permuted
                else:
                    X_train_permuted = self.Data_transformer_pipe.fit_transform(self.X_train_df.loc[:, selected_columns])
                    X_val_permuted = self.Data_transformer_pipe.transform(self.X_val_df.loc[:, selected_columns])
                    X_test_permuted = self.Data_transformer_pipe.transform(self.X_test_df.loc[:, selected_columns])
                yield X_train_permuted, X_val_permuted, X_test_permuted,selected_columns

    def only_predict(self,X, clusterer):
        labels = clusterer.predict(X) if hasattr(clusterer, 'predict') else clusterer.fit_predict(X)
        return labels



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.model_selection import train_test_split
from sklearn.model_selection import cross_val_score
from keras.api.models import Sequential
from keras.api.layers import Dense, BatchNormalization, Dropout
from keras.api.optimizers import Adam, SGD, RMSprop, Adadelta, Adagrad, Adamax, Nadam, Ftrl
from keras.api.callbacks import EarlyStopping, ModelCheckpoint
from math import floor
from sklearn.metrics import make_scorer, accuracy_score
from bayes_opt import BayesianOptimization
from sklearn.model_selection import StratifiedKFold
from keras.api.layers import LeakyReLU
LeakyReLU = LeakyReLU(alpha=0.1)
import warnings
warnings.filterwarnings('ignore')
pd.set_option("display.max_columns", None)
from scikeras.wrappers import KerasRegressor
from sklearn.metrics import mean_squared_error
from auto_ml_kinder import model_training_data_prep as mtdp
from keras.api.callbacks import LearningRateScheduler
from auto_ml_kinder.classes.ModelTrainingData import ModelTrainingData


class NeuralNetwork_Regression():
    data:ModelTrainingData
    model_params = None

    def __init__(self,data:ModelTrainingData) -> None:
        self.data = data

    def get_nn_model(self,neurons,total_columns,activation,normalization,hidden_layers,dropout,dropout_rate,optimizer):
        nn = Sequential()
        nn.add(Dense(neurons, input_dim=total_columns, activation=activation))

        if normalization > 0.5:
            nn.add(BatchNormalization())

        for _ in range(hidden_layers):
            hidden_neurons = round(neurons / (_ + 1))
            nn.add(Dense(hidden_neurons, activation=activation))

            if normalization > 0.5:
                nn.add(BatchNormalization())

            if dropout == 1:
                nn.add(Dropout(rate=dropout_rate))

        nn.add(Dense(1, activation='relu'))
        nn.compile(loss='mean_squared_error', optimizer=optimizer)
        return nn
    
    def lr_schedule(epoch, initial_lr):
        drop = 0.5
        epochs_drop = 10
        lr = initial_lr * (drop ** np.floor(epoch / epochs_drop))
        return lr

    # Create a LearningRateScheduler callback that uses the initial learning rate
    def create_lr_scheduler(self,initial_lr):
        return LearningRateScheduler(lambda epoch: self.lr_schedule(epoch, initial_lr))

# Create the LearningRateScheduler callback

    
    def nn_cl_bo2(self,neurons, activation, optimizer, learning_rate, batch_size, epochs, normalization, dropout, dropout_rate,hidden_layers):
        optimizerL = [
            'Adam'
            , 'RMSprop'
            , 'Adadelta'
            , 'Adagrad'
            , 'Adamax'
            , 'Nadam'
            , 'Ftrl'
            ]
        activationL = ['relu',LeakyReLU, 'selu','elu','sigmoid','softplus','softsign','tanh','exponential']
        neurons = round(neurons)
        activation = activationL[round(activation)]
        optimizer = optimizerL[round(optimizer)].lower()
        batch_size = round(batch_size)
        epochs = round(epochs)
        # layers1 = round(layers1)
        # layers2 = round(layers2)
        hidden_layers = round(hidden_layers)
        try:
            total_columns = len(self.data.X_train[0])
        except:
            total_columns = len(self.data.X_train.columns)

        self.print_things(activation, batch_size, epochs, hidden_layers, normalization, dropout, dropout_rate, neurons, optimizer, total_columns,seperator='||')

        def get_build_function():
            return self.get_nn_model(neurons,total_columns,activation,normalization,hidden_layers,dropout,dropout_rate,optimizer)
        
        lr_scheduler = self.create_lr_scheduler(learning_rate)
        mse_scorer = make_scorer(mean_squared_error, greater_is_better=True)
        es = EarlyStopping(monitor='loss', mode='min', verbose=0, patience=20)
        nn = KerasRegressor(build_fn=get_build_function, epochs=epochs, batch_size=batch_size, verbose=0)
        kfold = StratifiedKFold(n_splits=5, shuffle=True, random_state=123)
        score = cross_val_score(nn, self.data.X_train, self.data.Y_train, scoring=mse_scorer, cv=kfold, fit_params={'callbacks':[es,lr_scheduler]}).mean()
        import math
        score = round(score,1)
        return -((np.random.rand() * 1e30) if math.isnan(score) else score) # return large float if nan cz we cant show loss as 0
    
    def get_best_model(self,params):
        params_nn_ = params
        learning_rate = params_nn_['learning_rate']
        activationL = ['relu',LeakyReLU, 'selu','elu','sigmoid','softplus','softsign','tanh','exponential']
        activation = activationL[round(params_nn_['activation'])]
        batch_size = round(params_nn_['batch_size'])
        epochs = round(params_nn_['epochs'])
        normalization = round(params_nn_['normalization'])
        dropout = round(params_nn_['dropout'])
        dropout_rate = params_nn_['dropout_rate']
        neurons = round(params_nn_['neurons'])
        optimizerL = ['Adam', 'RMSprop', 'Adadelta', 'Adagrad', 'Adamax', 'Nadam', 'Ftrl','Adam']
        optimizer = optimizerL[round(params_nn_['optimizer'])].lower()
        hidden_layers = round(params_nn_['hidden_layers'])
        dropout = round(params_nn_['dropout'])
        
        try:
            total_columns = len(self.data.X_train[0])
        except:
            total_columns = len(self.data.X_train.columns)
        self.print_things(activation, batch_size, epochs, hidden_layers, normalization, dropout, dropout_rate, neurons, optimizer, total_columns)

        def get_build_function():
            return self.get_nn_model(neurons,total_columns,activation,normalization,hidden_layers,dropout,dropout_rate,optimizer)
        
        lr_scheduler = self.create_lr_scheduler(learning_rate)
        es = EarlyStopping(monitor='loss', mode='min', verbose=0, patience=5)
        nn_KerasRegressor = KerasRegressor(build_fn=get_build_function, epochs=epochs, batch_size=batch_size,verbose=0)
        nn_KerasRegressor.fit(self.data.X_train, self.data.Y_train, validation_data=(self.data.X_val, self.data.Y_val), verbose=1, callbacks=[es,lr_scheduler])
        return nn_KerasRegressor

    def print_things(self, activation, batch_size, epochs, hidden_layers, normalization, dropout, dropout_rate, neurons, optimizer, total_columns,seperator = '\n'):
        stringToPrint = f'Neurons: {neurons}{seperator}Total Columns:{total_columns}{seperator}Normalization:{normalization > .5}{seperator}Activation:{activation}{seperator}Total hidden layers:{hidden_layers}{seperator}Dropout used:{dropout == 1}{seperator}Dropout rate if used:{dropout_rate}{seperator}Optimizer:{optimizer}{seperator}Epochs:{epochs}{seperator}Batch_size:{batch_size}\n'
        print(stringToPrint)



from sklearn.ensemble import RandomForestRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.svm import SVR
from sklearn.linear_model import ElasticNet
from sklearn.ensemble import GradientBoostingRegressor
from enum import Enum
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from sklearn.linear_model import Lasso
from sklearn.neighbors import KNeighborsRegressor

class ModelAndParamRegression(Enum):
    Linear_Regression = LinearRegression
    Ridge_Regression = Ridge
    Lasso_Regression = Lasso
    ElasticNet_Regression = ElasticNet
    SVR_Regression = SVR
    DecisionTree_Regressor = DecisionTreeRegressor
    RandomForest_Regressor = RandomForestRegressor
    GradientBoosting_Regressor = GradientBoostingRegressor
    KNeighbors_Regressor = KNeighborsRegressor